% Loose Commitment in Medium- and Large-Scale Models (Debortoli, Maih, Nunes, 2010)
% Copyright: Debortoli, Maih, Nunes (2010). This version, November 2010
% Feel free to use and modify this code at your own risk.

% This file is the main engine to solve a problem under loose commitment.
% It implements the solution algorithm described in the paper.

% INPUTs: - Alag,A0,Alead,B: matrices describing the structural equations of the model (the notation is consistent with the paper
%         - W : matrix describing the objective function
%         - gam: probability of commitment
%         - beta: discount factor
%         - Reconvexify (unused) 
%         - H0: initial guess
%         - verbose: option to display errors on the screen (1=yes, 0=no)

% OUTPUTs: - H, G: matrices describing the law of motion
%          - iGAM0 : indicator for Lagrange Multipliers
%          - retcode: returns type of error

function [H,G,iGAM0,retcode]=SolveLooseCommitment(Alag,A0,Alead,B,W,gam,beta,ReConvexify,H0,verbose)
global options_

MaxIter=options_.MaxIterLC;
crit=options_.critLC;
qz_criterium=options_.qz_criteriumLC;

[mult_nbr,endo_nbr]=size(A0);
exo_nbr=size(B,2);
m=endo_nbr+mult_nbr;
y=1:endo_nbr;
l=endo_nbr+1:m;

if nargin==9
    verbose=0;
elseif nargin==8
    verbose=0;
    H0=zeros(m);
end
if isempty(H0)
    H0=zeros(m);
end

[GAM0,GAMlag,GAMv]=GetLooseCommitmentMatrices(H0);

Alag=sparse(Alag);
A0=sparse(A0);
Alead=sparse(Alead);
GAMlag=sparse(GAMlag);
GAMv=sparse(GAMv);

[H,iGAM0,retcode]=Iterate(H0);

switch retcode
    case 2 % nan
	    retcode=3;
	    retcode(2)=10000;
		if verbose
			disp([mfilename,':: NAN elements in the solution'])
		end
	case 1% maxiter
		if verbose
			disp([mfilename,':: Maximum Number of Iterations reached'])
		end
	case 0
	    BadEig=max(abs(eig(H)))>qz_criterium;
	    if BadEig
	        retcode=3;
	        retcode(2)=100*max(abs(eig(H)));
			if verbose
				disp([mfilename,':: Some eigenvalues greater than qz_criterium, Model potentially unstable'])
			end
	    end
end

if retcode(1)
    H=[];
    G=[];
else
    G=-iGAM0*GAMv;
end

    function [H,iGAM0,retcode]=Iterate(H0)
        iter=0;
        while 1
            iter=iter+1;
            GAM0=GetLooseCommitmentMatrices(H0);
            iGAM0=GAM0\eye(m);
            H=-iGAM0*GAMlag;
            [rcode,NQ]=CheckConvergence(H-H0,iter,MaxIter,crit);
            if rcode
                break
            else
                if verbose
                    disp(NQ)
                end
            end
            H0=H;
        end
        retcode=rcode-1;
    end

    function [GAM0,GAMlag,GAMv]=GetLooseCommitmentMatrices(H)
        GAM0=zeros(m);

        Hyy=H(y,y);
        Hyl=H(y,l);
        Hly=H(l,y);
        Hll=H(l,l);
        
        AleadH=Alead*Hyy;
        
        GAM0(y,y)=2*W+beta*Alag'*Hly; 
        GAM0(y,l)=transpose(gam*beta*Hll'*Alag+A0+(1-gam)*AleadH);
        
        GAM0(l,y)=A0+AleadH;
        GAM0(l,l)=gam*Alead*Hyl;
        
        if nargout>1
            GAMlag=zeros(m);
            if ReConvexify
                GAMlag(y,l)=gam/beta*Alead';
            else
                GAMlag(y,l)=(gam>0)/beta*Alead';
            end
            GAMlag(l,y)=Alag;
            if nargout==3
                GAMv=zeros(m,exo_nbr);
                GAMv(l,:)=B;
            end
        end
    end
end

function [rcode,NQ]=CheckConvergence(Q,iter,MaxIter,crit)

NQ=max(max(abs(Q)));% norm(Q); seems too costly
if isnan(NQ)
    rcode=3;
elseif iter>MaxIter;
    rcode=2;
elseif NQ<crit
    rcode=1;
else
    rcode=0;
end
end